     //////////////////////////////////////////////
    // THE COMPARATIVE CODE-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad                   ///
  //                                           ////
 // CCUS_FMA.java (2002:07:29 15:53)          /////
///////////////////////////////////////////////////
// Methods of CCUS_FMA:                      //////
// - indent                                  //////
// - matchFrames                             /////
// - matchFillers(Object,Object)             ////
// - matchFillers(Object,Object,Slot,Object) ///
// - processDifference                       //
//////////////////////////////////////////////

import java.io.*;
import java.util.*;
import jex.*;

/*****************************************************************************
** The class `CCUS_FMA' implements the basic Frame Matching Algorithm for the
** Comparative Constraint-Understanding System (this is much like the case
** matching algorithm in class Case).
** This class is not meant to be instantiated.
*****************************************************************************/

public class CCUS_FMA {
	public static int _numberOfComparisons;
	public static int _numberOfHLCComparisons;
	public static int _numberOfDifferences;

	private static short _indent= -1;

	/********************************************************************/
	// CCUS_FMA.indent : void -> void
	/*********************************************************************
	** Print _indent number of '>'s at the beginning of a line.
	*********************************************************************/

	public static void indent () {
		for (short i= 0; i < _indent; i++)
			System.out.print(".");
	}

	/********************************************************************/
	// CCUS_FMA.matchFrames : (Frame x Frame x byte x Object) -> boolean
	/*********************************************************************
	** Compares a frame in one model with a frame in the other. The
	** Boolean value it returns is of secondary importance to the
	** correspondences it infers and the differences it finds. A
	** precondition is that the frames have already been inferred to
	** correspond, but this method could return false if the
	** correspondence is only plausible and has to be retracted.
	** @param frame1 The frame to compare in the first model.
	** @param frame2 The frame to compare in the second model.
	** @param mode Which mode it's comparing in (CORRESPONDENCE or
	**             PLAUSIBLE_CORRESPONDENCE).
	** @param subsForSelf A value to substitute for SELF (optional - can
	**                    be null).
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return True, unless the comparison fails.
	*********************************************************************/

	public static boolean matchFrames (Frame frame1, Frame frame2,
	                                   byte mode, Object subsForSelf,
	                                   EntityPair entityPair)
	{
		_indent++;

		///////////////////////////////////////////////
		// If the frames have been compared already, //
		// then cut to avoid infinite recursion.     //
		///////////////////////////////////////////////

		if (Correspondences.correspondToEachOther(frame1,frame2)) {
			_indent--;
		 	return true;
		}
		if (Correspondences.alreadyComparedAndFailed(frame1,frame2)) {
			_indent--;
			return false;
		}

		/////////////////////////////////////////////////////
		// Record correspondence between the two frames    //
		// (must be done now to avoid infinite recursion). //
		/////////////////////////////////////////////////////

		Correspondences.recordCorrespondence(frame1, frame2, mode);

		//////////////////////////////////////////////////////////
		// If the frames are entities, create a new EntityPair. //
		//////////////////////////////////////////////////////////

		boolean areEntities= false;

		if (CCUS._generateSummary
		 && frame1.getFrameClass().getName().equals("entity_type")
		 && frame2.getFrameClass().getName().equals("entity_type"))
		{
			entityPair= new EntityPair(frame1, frame2);
			areEntities= true;
		}

		////////////////////////////////////////////////////////
		// Match each slot filled in the first frame with the //
		// corresponding slot in the second frame (only those //
		// slots which were set directly in the first frame,  //
		// and ignoring any slots marked with --IGNORE).      //
		////////////////////////////////////////////////////////

		Hashtable slotsAlreadyMatched= new Hashtable();

		Case seCase= null; // will retrieve a case at first difference

		Object substituteForSelf= null;
		try {
			Object sfs1= frame1.getSlotFiller
			             ("substitute_for_self");
			Object sfs2= frame2.getSlotFiller
			             ("substitute_for_self");
			if (sfs1 != null && sfs2 != null) {
				System.out.println("Warning: CCUS_FMA.matchFr"
				+ "ames: " + "sfs1 != null && sfs2 != null");
			}
			if (sfs1 != null) {
				substituteForSelf= sfs1;
			}
			if (sfs2 != null) {
				substituteForSelf= sfs2;
			}
			if (substituteForSelf != null && subsForSelf != null){
				System.err.println("CCUS_FMA.matchFrames: "
				+ "substituteForSelf != null but "
				+ "subsForSelf != null");
				System.exit(1);			
			}
		} catch (InvalidSlot is) {}

		if (subsForSelf != null) {
			substituteForSelf= subsForSelf;
		}

		ListIterator slots1= frame1.getSlots();
		while (slots1.hasNext()) {
			Slot slot= (Slot)slots1.next();
			////////////////////////////////
			// Ignore inappropriate slots //
			////////////////////////////////
			if (slot.getIgnore())
				continue;

			boolean m= true;

			///////////////////////////////////////
			// Deal with slot-filler differences //
			///////////////////////////////////////
			if (frame1.isSlotSetDirectly(slot)
			 && ! (m= matchFillers(frame1.getSlotFiller(slot),
			                       frame2.getSlotFiller(slot),
			                       slot, mode, substituteForSelf,
			                       entityPair)))
			{
				try {
					seCase= processDifference(frame1,
					  frame2,slot,mode,seCase,entityPair);
				} catch (FailMatch fm) {
					Correspondences.retractCorrespondence
					                (frame1, frame2);
					_indent--;
					return false;
				}
			}
			else if (CCUS._reportNonDifferences) {
				indent();
				System.out.println("No difference in slot \""
				+ slot + "\" (frame1)");
			}
			slotsAlreadyMatched.put(slot, new Boolean(m));
		}

		////////////////////////////////////////////////////////////
		// Match each slot filled in the second frame with the    //
		// corresponding slot in the first frame (only those      //
		// slots which were set directly in the second frame,     //
		// and ignoring any slots marked with --IGNORE).          //
		// Do not repeat slot-matches which were performed above. //
		////////////////////////////////////////////////////////////

		ListIterator slots2= frame2.getSlots();
		while (slots2.hasNext()) {
			Slot slot= (Slot)slots2.next();
			if (slot.getIgnore()
			    || slotsAlreadyMatched.get(slot) != null)
			{
				continue; // skip this slot
			}

			///////////////////////////////////////
			// Deal with slot-filler differences //
			///////////////////////////////////////
			if (frame2.isSlotSetDirectly(slot)
			 && ! matchFillers(frame1.getSlotFiller(slot),
			                   frame2.getSlotFiller(slot),
			                   slot, mode, substituteForSelf,
			                   entityPair))
			{
				try {
					seCase= processDifference(frame1,
					  frame2,slot,mode,seCase,entityPair);
				} catch (FailMatch fm) {
					Correspondences.retractCorrespondence
					                (frame1, frame2);
					_indent--;
					return false;
				}
			}
			else if (CCUS._reportNonDifferences) {
				indent();
				System.out.println("No difference in slot \""
				+ slot + "\" (frame2)");
			}
		}

		////////////////////////////////////////////
		// Match succeeds - unless it failed! ;-) //
		////////////////////////////////////////////
		if (mode == CorrespondenceHeuristics.PLAUSIBLE_CORRESPONDENCE)
		{
			indent();
			System.out.println("Correspondence between " + frame1
			+ " and " + frame2 + " confirmed.");
		}

		if (CCUS._generateSummary && entityPair != null
		 && ! areEntities)
		{
			entityPair.addCorrespondence(frame1, frame2);
		}
//else if (! areEntities) System.out.println("@ Correspondence between " + frame1 + " and " + frame2 + " not added to summary (entityPair == " + entityPair + ")");

		//////////////////////////////////////////////////////
		// If the frames' "covers" slots are set,           //
		// record correspondence between the covered frames //
		//////////////////////////////////////////////////////

		Frame cover1= (Frame)frame1.gsf("covers");
		Frame cover2= (Frame)frame2.gsf("covers");
		if (cover1 != null && cover2 != null
		 && ! Correspondences.correspondToEachOther(cover1,cover2))
		{
			Correspondences.recordCorrespondence(cover1, cover2,
			  CorrespondenceHeuristics.CORRESPONDENCE);
		}

		_indent--; return true;
	}

	/********************************************************************/
	// CCUS_FMA.matchFillers : (Object x Object) -> boolean
	/*********************************************************************
	** Matches two slot fillers - one from the first frame, one from the
	** second. This method just calls matchFillers(filler1, filler2, null,
	** PLAUSIBLE_CORRESPONDENCE, null, entityPair).
	** @param filler1 A slot filler from the first frame.
	** @param filler2 A (corresponding) slot filler from the second frame.
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return Whether the two slot fillers match.
	*********************************************************************/

	public static boolean matchFillers (Object filler1, Object filler2,
	                                    EntityPair entityPair)
	{
		return matchFillers(filler1, filler2, null,
		       CorrespondenceHeuristics.PLAUSIBLE_CORRESPONDENCE,
		       null, entityPair);
	}

	/********************************************************************/
	// CCUS_FMA.matchFillers : (Object x Object x Slot x byte x Object)
	//                         -> boolean
	/*********************************************************************
	** Matches two slot fillers - one from the first frame, one from the
	** second.
	** @param filler1 A slot filler from the first frame.
	** @param filler2 A (corresponding) slot filler from the second frame.
	** @param slot The slot whose fillers are being matched (optional -
	**             can be null). This is for the information of matching
	**             procedures that take the type of the slot into account.
	** @param mode Which mode it's comparing in (CORRESPONDENCE or
	**             PLAUSIBLE_CORRESPONDENCE).
	** @param subsForSelf A value to substitute for SELF (optional - can
	**                    be null).
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return Whether the two slot fillers match.
	*********************************************************************/

	public static boolean matchFillers (Object filler1, Object filler2,
	                                    Slot slot, byte mode,
	                                    Object subsForSelf,
	                                    EntityPair entityPair)
	{
		_numberOfComparisons++;

		//////////////////////////////////////
		// If one slot filler is null then  //
		// they must both be null to match. //
		//////////////////////////////////////
		if (filler1 == null || filler2 == null)
			return (filler2 == null && filler1 == null);

		/////////////////////////////////////////////////////////
		// Match fails if second slot filler is not Java-type- //
		// compatible with first slot filler.                  //
		/////////////////////////////////////////////////////////

		if (! filler1.getClass().isInstance(filler2)
		 && ! filler2.getClass().isInstance(filler1))
			return false;

		//////////////////////////
		// Substitute for SELF. //
		//////////////////////////

		if (subsForSelf != null) {
			boolean self1= filler1 instanceof Frame
			 && ((Frame)filler1).getFrameClass()
			    == FrameClass.lookup("built_in_constant")
			 && ((DefinedTypeInstance)
			     ((Frame)filler1).gsf("built_in_value")).
			     toString().equals("self_constant");

			boolean self2= filler2 instanceof Frame
			 && ((Frame)filler2).getFrameClass()
			    == FrameClass.lookup("built_in_constant")
			 && ((DefinedTypeInstance)
			     ((Frame)filler2).gsf("built_in_value")).
			     toString().equals("self_constant");

			if (self1 && !self2) {
				filler1= subsForSelf;
			}
			if (self2 && !self1) {
				filler2= subsForSelf;
			}
		}

		/////////////////////////////////////////////////
		// If they're frames, match them recursively,  //
		// provided their correspondence is sanctioned //
		// by the correspondence heuristics.           //
		/////////////////////////////////////////////////

		if (filler1 instanceof Frame && filler2 instanceof Frame) {
			if (FrameClass.lookup("higher_level_constraint").
		            subsumes(((Frame)filler1).getFrameClass()))
		        {
		        	_numberOfHLCComparisons++;
		        }
		        mode= CorrespondenceHeuristics.sanctionCorrespondence
			       ((Frame)filler1, (Frame)filler2, entityPair);
			if (mode == CorrespondenceHeuristics.CORRESPONDENCE
			 || mode == CorrespondenceHeuristics.
			             PLAUSIBLE_CORRESPONDENCE)
			{
				return matchFrames
				       ((Frame)filler1, (Frame)filler2, mode,
				        subsForSelf, entityPair);
			}
		}

		//////////////////////////////////////////
		// If they're defined type instances,   //
		// compare type then underlying values. //
		//////////////////////////////////////////

		if (filler1 instanceof DefinedTypeInstance
		 && filler2 instanceof DefinedTypeInstance)
		{
			return (((DefinedTypeInstance)filler1).
			         getDefinedType()
			     == ((DefinedTypeInstance)filler2).
			         getDefinedType())
			&&
			matchFillers
			(((DefinedTypeInstance)filler1).getValue(),
			 ((DefinedTypeInstance)filler2).getValue(), null,
			 mode, subsForSelf, entityPair);
		}

		//////////////////
		// Aggregations //
		//////////////////

		if (filler1 instanceof Vector && filler2 instanceof Vector) {
			return AggregationMatchingPolicies.
			       matchVectors((Vector)filler1, (Vector)filler2,
			                    mode, (slot == null)
			                          ? null
			                          : slot.getType(),
			                    subsForSelf, entityPair);
		}

		/////////////////////////////////////
		// By default, call Object.equals. //
		/////////////////////////////////////

		return filler1.equals(filler2);
	}

	/********************************************************************/
	// CCUS_FMA.processDifference : (Frame x Frame x Slot x byte x Case)
	//                              -> Case
	/*********************************************************************
	** Deal with differences in slot-fillers found during matchFrames.
	** Report the difference, and if matching in PLAUSIBLE_CORRESPONDENCE
	** mode, use a semantic-equivalence/strength case to decide whether to
	** fail the match, or which frame is stronger.
	** @param frame1 The frame being matched in the first model.
	** @param frame2 The frame being matched in the second model.
	** @param slot The slot for which a difference has been found.
	** @param mode Which mode it's matching in (CORRESPONDENCE or
	**             PLAUSIBLE_CORRESPONDENCE).
	** @param seCase The semantic-equivalence/strength case that pertains
	**               to the current frame-match, if there is one.
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return seCase (if a semantic case is retrieved, the caller
	**         matchFrames needs a reference to it).
	** @exception FailMatch Tells the caller matchFrames to fail the match
	**                      if the difference means that the frames are
	**                      semantically incomparable. It's the caller's
	**                      responsibility to retract the correspondence.
	*********************************************************************/

	private static Case processDifference (Frame frame1, Frame frame2,
	                                       Slot slot, byte mode,
	                                       Case seCase,
	                                       EntityPair entityPair)
	                    throws FailMatch
	{	
		_numberOfDifferences++;

		Object filler1= frame1.getSlotFiller(slot);
		Object filler2= frame2.getSlotFiller(slot);

		indent();
		System.out.println("Difference in slot \"" + slot + "\":-");
		indent();
		System.out.println("- " + frame1 + "." + slot + " == "
		+ filler1);
		indent();
		System.out.println("- " + frame2 + "." + slot + " == "
		+ filler2);
		//indent(); System.out.println("\n");

		if (CCUS._generateSummary && entityPair != null
		 && mode == CorrespondenceHeuristics.CORRESPONDENCE
		 && !(filler1 instanceof UnorderedVector &&
		      filler2 instanceof UnorderedVector))
		{
			entityPair.addDifference(slot, frame1, filler1,
			                               frame2, filler2);
		}

		/////////////////////////////////////////////
		// Use semantic-equivalence/strength cases //
		// in PLAUSIBLE_CORRESPONDENCE mode.       //
		/////////////////////////////////////////////
		if (mode == CorrespondenceHeuristics.PLAUSIBLE_CORRESPONDENCE)
		{
			try {
				if (! CCUS._semanticCases) {
					throw new FailMatch();
				}
				return SemanticStrength.useSemanticCase
				       (frame1,frame2,slot,seCase,entityPair);
			} catch (FailMatch fm) {
				if (CCUS._extractHLCs) {
					HLC_CCUS.extractHLCs
					   (frame1, frame2, slot, entityPair);
				}
				if (filler1 instanceof Frame
				 && filler2 instanceof Frame)
				{
					if (Correspondences.
					    correspondToEachOther
					    ((Frame)filler1,(Frame)filler2))
					{
						return seCase;
					}
				}
				throw fm;
			}
		}
		/////////////////////////
		// CORRESPONDENCE mode //
		/////////////////////////
		if (CCUS._extractHLCs) {
			HLC_CCUS.extractHLCs(frame1,frame2,slot,entityPair);
		}
		return seCase;
	}
}
