      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // CaseLibrary.java (2000:05:04 16:21) //////
//////////////////////////////////////////////
// Methods of CaseLibrary:             ///////
// - retrieve                          ///////
// - retrieveOneMatchingCase           ///////
// - static selectMostSpecific         ///////
// - insert                            //////
// - index                             /////
// - static load                       ////
// - static save                       ///
// - print                             //
////////////////////////////////////////

import java.io.*;
import java.util.*;

/*****************************************************************************
** The class `CaseLibrary' represents a case library. It encapsulates an
** organisation of cases within the library, allowing new cases to be added
** and existing cases to be retrieved.
*****************************************************************************/

public class CaseLibrary implements Serializable {
	private Hashtable/*FrameClass->Vector<Case>*/ _organ= new Hashtable();

	/********************************************************************/
	// CaseLibrary.retrieve : Frame -> Vector<Case>
	/*********************************************************************
	** Retrieves those cases from the case library that are indexed by the
	** frame class of the prompt frame, without full matching.
	** @param prompt A frame to use as a retrieval prompt.
	** @return The set of cases which the prompt reminds the CBR system
	**         of.
	*********************************************************************/

	public Vector retrieve (Frame prompt) {
		Vector vector= (Vector)_organ.get(prompt.getFrameClass());
		if (vector == null)
			vector= new Vector();
		return vector;
	}

	/********************************************************************/
	// CaseLibrary.retrieveOneMatchingCase : Frame -> Case
	/*********************************************************************
	** Retrieves one case from the case library that fully matches the
	** prompt frame. If it finds such a case, it returns it immediately,
	** without continuing to see if there are any more (there shouldn't
	** be, by construction).
	** @param prompt A frame to use as a retrieval prompt.
	** @return One case that fully matches the prompt frame, or null if
	**         no such case is retrieved.
	*********************************************************************/

	public Case retrieveOneMatchingCase (Frame prompt) {
		Vector retrievedCases= retrieve(prompt);
		ListIterator rCases= retrievedCases.listIterator();
		while (rCases.hasNext()) {
			Case c= (Case)rCases.next();
			if (c.match(prompt)) {
				return c;
			}
		}
		return null;
	}

	/********************************************************************/
	// static CaseLibrary.selectMostSpecific : Vector<Case> -> Case
	/*********************************************************************
	** Given a set of cases, select the one which is the most specific -
	** i.e. whose `problem' frame class is a subclass of all the others'.
	** @param cases A non-empty set of cases.
	** @return The most specific case in the set.
	*********************************************************************/

	public static Case selectMostSpecific (Vector cases) {
		ListIterator caseEnum= cases.listIterator();
		Case mostSpecific= (Case)caseEnum.next();
		FrameClass mSPFC= mostSpecific.getProblemFrameClass();
		while (caseEnum.hasNext()) {
			Case currentCase= (Case)caseEnum.next();
			FrameClass cCPFC= currentCase.getProblemFrameClass();
			mSPFC= mostSpecific.getProblemFrameClass();

			if (mSPFC.subsumes(cCPFC)) {
				if (mSPFC == cCPFC) {
					System.err.println("CaseLibrary.selec"
					+ "tMostSpecific: two cases have the "
					+ "same problem frame class - can't "
					+ "choose between them!");
					System.exit(1);
				}
				mostSpecific= currentCase; mSPFC= cCPFC;
			}
		}
		return mostSpecific;
	}

	/********************************************************************/
	// CaseLibrary.insert : Case -> void
	/*********************************************************************
	** Adds a new case in the case library. Indexes it by the frame class
	** of the frame in its "problem" slot, and by all the subclasses of
	** that frame class (so that it will be retrieved for prompt frames
	** which are instances of the subclasses).
	** @param case The new case to store in the case library.
	*********************************************************************/

	public void insert (Case newCase) {
		index(newCase, newCase.getProblemFrameClass());
	}

	/********************************************************************/
	// CaseLibrary.index : (Case x FrameClass) -> void
	/*********************************************************************
	** Auxiliary function to index a case by a frame class (when inserting
	** the case into the case library). The function indexes the case by
	** the frame class itself (unless it is abstract), and recurses for
	** each of the frame class's subclasses.
	** @param kase The case to index (`case' is a reserved word in Java).
	** @param frameClass The frame class to index by.
	*********************************************************************/

	private void index (Case kase, FrameClass frameClass) {
		if (! frameClass.isAbstract()) {
			Vector vector= (Vector)_organ.get(frameClass);
			if (vector == null)
				_organ.put(frameClass, vector= new Vector());
			vector.addElement(kase);
		}
		ListIterator subclasses= frameClass.getSubClasses().
		                                    listIterator();
		while (subclasses.hasNext()) {
			index(kase, (FrameClass)subclasses.next());
		}
	}

	/********************************************************************/
	// static CaseLibrary.load : String -> void
	/*********************************************************************
	** Loads the case library from a binary file.
	** @param filename The filename of the case library binary.
	** @exception FileNotFoundException If the file is not found.
	*********************************************************************/

	public static void load (String filename)
	                        throws FileNotFoundException
	{
		try {
			FileInputStream f= new FileInputStream(filename);
			ObjectInput s= new ObjectInputStream(f);

			FrameClass.readClass(s);
			Slot.readClass(s);
			DefinedType.readClass(s);
			FrameTemplate.readClass(s);
			JexToFrames.readClass(s);
			ExpressCUS.readClass(s);
			JavaCUS.readClass(s);
			f.close();
		} catch (FileNotFoundException fnfe) {
			throw fnfe;
		} catch (IOException ioe) {
			System.err.println(ioe);
			System.exit(1);
		}
	}

	/********************************************************************/
	// CaseLibrary.save : String -> void
	/*********************************************************************
	** Saves the case library to a binary file.
	** @param filename The filename of the case library binary.
	*********************************************************************/

	public static void save (String filename) {
		try {
			FileOutputStream f= new FileOutputStream(filename);
			ObjectOutput s= new ObjectOutputStream(f);

			FrameClass.writeClass(s);
			Slot.writeClass(s);
			DefinedType.writeClass(s);
			FrameTemplate.writeClass(s);
			JexToFrames.writeClass(s);
			ExpressCUS.writeClass(s);
			JavaCUS.writeClass(s);
			f.close();
		} catch (IOException x) {
			System.err.println(x);
			System.exit(1);
		}		
	}

	/********************************************************************/
	// CaseLibrary.print : void -> void
	/*********************************************************************
	** Prints the contents of the case library, in the form of a list of
	** cases (frame reference and filename for each case) for each frame
	** class in the library's index.
	*********************************************************************/

	public void print () {
		System.out.println("++++++++++++++++++++++++++++++++++++++++"
		+ "++++++++++++++++++++++++++++++++++++++++");

		Enumeration keys= _organ.keys();
		while (keys.hasMoreElements()) {
			FrameClass fc= (FrameClass)keys.nextElement();
			System.out.println(fc + ":-");
			ListIterator cases= ((Vector)_organ.get(fc)).
			                    listIterator();
			while (cases.hasNext()) {
				Case c= (Case)cases.next();
				System.out.println("\t" + c.getCaseFrame()
				+" \""+c.getCaseFrame().gsf("filename")+"\"");
			}
			System.out.println();
		}

		System.out.println("++++++++++++++++++++++++++++++++++++++++"
		+ "++++++++++++++++++++++++++++++++++++++++");
	}
}
