     /////////////////////////////////////////////////////
    // THE COMPARATIVE CODE-UNDERSTANDING SYSTEM        //
   // Written by Andrew Broad                          ///
  //                                                  ////
 // CorrespondenceHeuristics.java (2002:01:10 13:07) /////
//////////////////////////////////////////////////////////
// Methods of CorrespondenceHeuristics:             //////
// - subexpressionCaseLibrary                       //////
// - sanctionCorrespondence                         //////
// - frameClassEquivalence                          /////
// - nameEquivalence                                ////
// - synonyms                                       ///
// - comparableFrameClasses                         //
/////////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import jex.*;

import jwnl.data.*;
import jwnl.data.list.pointer_target.*;
import jwnl.data.relationship.*;
import jwnl.dictionary.DictionaryDatabase;
import jwnl.dictionary.hashtable.HashtableBackedDictionary;

/*****************************************************************************
** The class `CorrespondenceHeuristics' implements various correspondence
** heuristics used to infer correspondences between frames.
*****************************************************************************/

public class CorrespondenceHeuristics {
	public final static byte NO_CORRESPONDENCE= 0;
	public final static byte PLAUSIBLE_CORRESPONDENCE= 1;
	public final static byte CORRESPONDENCE= 2;

	public static int _nameThesaurusLookups;
	public static int _frameClassThesaurusLookups;

	/********************************************************************/
	// static CorrespondenceHeuristics.subexpressionCaseLibrary :
	//                                 void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the subexpression case library of the
	** EXPRESS CUS (if the CCUS is extended to other languages
	** (e.g. Java), it will have to be parameterised with respect to
	** subexpression case library).
	** @return The subexpression case library.
	*********************************************************************/

	private static CaseLibrary subexpressionCaseLibrary () {
		return ExpressCUS.getSubexpressionCaseLibrary();
	}

	/********************************************************************/
	// static CorrespondenceHeuristics.sanctionCorrespondence :
	//        (Frame x Frame x EntityPair) -> byte
	/*********************************************************************
	** Returns CORRESPONDENCE if two frames definitely correspond
	** (according to an appropriate correspondence heuristic),
	** NO_CORRESPONDENCE if they definitely don't correspond,
	** and PLAUSIBLE_CORRESPONDENCE if they might correspond (but might
	** later be found not to).
	** @param frame1 The first frame.
	** @param frame2 The second frame.
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return One of CORRESPONDENCE, PLAUSIBLE_CORRESPONDENCE and
	**         NO_CORRESPONDENCE.
	*********************************************************************/

	public static byte sanctionCorrespondence (Frame frame1, Frame frame2,
	                                           EntityPair entityPair)
	{
		///////////////////////////////////////////
		// Test frame class equivalence first... //
		///////////////////////////////////////////
		if (frameClassEquivalence(frame1, frame2, entityPair)
		    == NO_CORRESPONDENCE)
			return NO_CORRESPONDENCE;

		///////////////////////////////
		// ...then name equivalence. //
		///////////////////////////////
		return nameEquivalence(frame1, frame2);
	}

	/********************************************************************/
	// static CorrespondenceHeuristics.frameClassEquivalence :
	//        (Frame x Frame x EntityPair) -> byte
	/*********************************************************************
	** Returns CORRESPONDENCE if two frames have the same frame class,
	** NO_CORRESPONDENCE if they have different frame classes.
	** @param frame1 The first frame.
	** @param frame2 The second frame.
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return One of CORRESPONDENCE and NO_CORRESPONDENCE.
	*********************************************************************/

	private static byte frameClassEquivalence (Frame frame1, Frame frame2,
	                                           EntityPair entityPair)
	{
		if (frame1.getFrameClass() == frame2.getFrameClass()) {
			return CORRESPONDENCE;
		} else {
			if (CCUS._frameClassThesaurus
			    && comparableFrameClasses(frame1.getFrameClass(),
			                              frame2.getFrameClass()))
			{
				return CORRESPONDENCE;
			} else {
				if (CCUS._subexpressionCases) {
					subexpressions(frame1, frame2, frame1,
					               entityPair);
					subexpressions(frame1,frame2,frame2,
					               entityPair);
				}
				return NO_CORRESPONDENCE;
			}
		}
	}

	/********************************************************************/
	// static CorrespondenceHeuristics.subexpressions :
	//        (Frame x Frame x Frame) -> void
	/*********************************************************************
	** If subexpression cases suggest it, match one of the frames to each
	** of the subexpressions of the other frame.
	** @param frame1 Frame from the first model.
	** @param frame2 Frame from the second model.
	** @param chosen Which of frame1 or frame2 is to have its
	**               subexpressions considered.
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	*********************************************************************/

	private static void subexpressions (Frame frame1, Frame frame2,
	                                  Frame chosen, EntityPair entityPair)
	{
		Vector subexpressionCases= subexpressionCaseLibrary().
		                           retrieve(chosen);

		if (subexpressionCases.size() >= 1) {
			Case subexpressionCase= CaseLibrary.
			  selectMostSpecific(subexpressionCases);
			if (subexpressionCase.match(chosen)) {
				Vector submits= subexpressionCase.extract();
				if (submits.size() > 1) {
					System.err.println("You'll have to ex"
					+ "tend CorrespondenceHeuristics.sube"
					+ "xpressions if you want to have mul"
					+ "tiple solution frames in subexpres"
					+ "sion cases! ;-)");
					System.exit(1);
				}
				Frame submit= (Frame)submits.elementAt(0);

				Vector submissionFrames
				= (Vector)submit.gsf("frames");
				Enumeration sf= submissionFrames.elements();
				Object item;
				while (sf.hasMoreElements()) {
					item= sf.nextElement();
					if (item instanceof Frame) {
						if (chosen == frame1) {
							CCUS_FMA.matchFillers
							((Frame)item, frame2,
							null, PLAUSIBLE_CORRESPONDENCE,
							null, entityPair);
						} else {
							CCUS_FMA.matchFillers
							(frame1, (Frame)item,
							null, PLAUSIBLE_CORRESPONDENCE,
							null, entityPair);
						}
					} else if (item instanceof Vector) {
						Enumeration f
						= ((Vector)item).elements();
						while (f.hasMoreElements()) {
							Object frame
							= f.nextElement();
							if (frame instanceof
							    Frame)
							{
							 if (chosen==frame1) {
							  CCUS_FMA.
							  matchFillers((Frame)
							  frame, frame2, null,
							  PLAUSIBLE_CORRESPONDENCE,
							  null, entityPair);
							 } else {
							  CCUS_FMA.
							  matchFillers(frame1,
							  (Frame)frame, null,
							  PLAUSIBLE_CORRESPONDENCE,
							  null, entityPair);
							 }
							} else
							 System.err.println(
							 "CorrespondenceHeur"+
							 "istics: Warning: f"+
							 "rame is not a Fram"+
							 "e - not submitted");
						}// elihw
					}// fi
				}// elihw
				Frame.remove(submit);
			} else {
				System.err.println("CorrespondenceHeuristics."
				+ "subexpression: subexpression case failed t"
				+ "o match!");
				System.exit(1);
			}// fi (subexpressionCase.match(chosen))
		}// fi (subexpressionCases.size() >= 1)
	}

	/********************************************************************/
	// static CorrespondenceHeuristics.nameEquivalence :
	//        (Frame x Frame) -> byte
	/*********************************************************************
	** Returns CORRESPONDENCE if two frames correspond according to name
	** equivalence, NO_CORRESPONDENCE if they do not, and
	** PLAUSIBLE_CORRESPONDENCE if they have no names.
	** @param frame1 The first frame.
	** @param frame2 The second frame.
	** @return One of CORRESPONDENCE, PLAUSIBLE_CORRESPONDENCE and
	**         NO_CORRESPONDENCE.
	*********************************************************************/

	private static byte nameEquivalence (Frame frame1, Frame frame2)
	{
		try {
			String frame1name= frame1.getSlotFiller("name").
			                          toString();
			String frame2name= frame2.getSlotFiller("name").
			                          toString();

			if (frame1name != null && frame2name != null
			    && frame1name.equalsIgnoreCase(frame2name))
			{
				return CORRESPONDENCE;
			} else {
				return CCUS._nameThesaurus
				       && synonyms(frame1name, frame2name)
				       ? CORRESPONDENCE
				       : NO_CORRESPONDENCE;
			}
		} catch (InvalidSlot is) {
			return PLAUSIBLE_CORRESPONDENCE;
		}
	}

	/********************************************************************/
	// static CorrespondenceHeuristics.synonyms :
	//        (String x String) -> boolean
	/*********************************************************************
	** Given two names, return true if they are synonyms in the thesaurus,
	** false otherwise. Before the names are looked up in the thesaurus,
	** they are stripped of underscores and converted to lower case.
	** @param name1 The first name.
	** @param name2 The second name.
	** @return true iff name1 and name2 are synonyms.
	*********************************************************************/

	private static boolean synonyms (String name1, String name2) {
		_nameThesaurusLookups++;

		IndexWord iw1= CCUS._db.lookupWord(POS.NOUN,
		                                   stripUnderscores(name1).
		                                   toLowerCase());
		if (iw1 == null) return false;
		IndexWord iw2= CCUS._db.lookupWord(POS.NOUN,
		                                   stripUnderscores(name2).
		                                   toLowerCase());
		if (iw2 == null) return false;
		RelationshipList rl= Relationship.findSynonymRelationships
		                                  (iw1, 1, iw2, 1);
		return (rl.size() > 0);
	}

	/********************************************************************/
	// static CorrespondenceHeuristics.comparableFrameClasses :
	//        (FrameClass x FrameClass) -> boolean
	/*********************************************************************
	** Given two frame classes, return true if they are comparable
	** according to the frame-class thesaurus, false otherwise. Before
	** the frame-class names are looked up in the thesaurus, they are
	** stripped of underscores (they are assumed to be lower-case).
	** @param name1 The first name.
	** @param name2 The second name.
	** @return true iff name1 and name2 are synonyms.
	*********************************************************************/

	private static boolean comparableFrameClasses (FrameClass frameClass1,
	                                               FrameClass frameClass2)
	{
		_frameClassThesaurusLookups++;
		IndexWord iw1= CCUS._db.lookupWord(POS.ADVERB,
		               stripUnderscores(frameClass1.getName()));
		if (iw1 == null) return false;
		IndexWord iw2= CCUS._db.lookupWord(POS.ADVERB,
		               stripUnderscores(frameClass2.getName()));
		if (iw2 == null) return false;
		RelationshipList rl= Relationship.findSynonymRelationships
		                                  (iw1, 1, iw2, 1);
		return (rl.size() > 0);
	}

	/********************************************************************/
	// static CorrespondenceHeuristics.stripUnderscores : String -> String
	/*********************************************************************
	** Removes underscores (`_`) from a string,
	** e.g. "entity_type" -> "entitytype".
	** @param s The string from which to remove underscores.
	** @return A string which is a copy of s with underscores removed.
	*********************************************************************/
 
 	public static String stripUnderscores (String s) {
		String t= ""; char c;
		for (int i= 0; i< s.length(); i++) {
			if ((c= s.charAt(i)) != '_') {
				t+= c;
			}
		}
		return t;
	}
}
