     //////////////////////////////////////////////
    // THE COMPARATIVE CODE-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad                   ///
  //                                           ////
 // Correspondences.java (2002:12:02 13:39)   /////
///////////////////////////////////////////////////
// Methods of Correspondences:               //////
// - recordCorrespondence                    /////
// - retractCorrespondence                   ////
// - correspondToEachOther                   ///
// - correspondenceFrame                     //
//////////////////////////////////////////////

import java.util.*;
import java.io.*;

import jex.*;

/*****************************************************************************
** The class `Correspondences' contains the methods to record a correspondence
** between two frames, to retract a correspondence between two frames, and to
** test whether two frames correspond to each other.
** This class is not meant to be instantiated.
*****************************************************************************/

public class Correspondences {
	public static int _numberOfCorrespondences;
	public static int _numberOfHLCCorrespondences;
	public static Hashtable/*FrameClass->Integer*/
		_numberOfCorrespondencesHashtable= new Hashtable();

	private static Hashtable _alreadyComparedAndFailed= new Hashtable();

	/********************************************************************/
	// static Correspondences.recordCorrespondence :
	//                        (Frame x Frame x byte) -> void
	/*********************************************************************
	** Records a correspondence between two frames by setting their
	** "corresponds_to" slots to each other. However, if it's a plausible
	** correspondence, creates a frame to represent the correspondences,
	** and sets the two frames' "corresponds_to" slots to point to the
	** correspondence frame. The correspondence frame may be tagged with
	** extra information about the correspondence later.
	** @param frame1 The first frame.
	** @param frame2 The second frame.
	** @param mode CORRESPONDENCE or PLAUSIBLE_CORRESPONDENCE.
	*********************************************************************/

	public static void recordCorrespondence (Frame frame1, Frame frame2,
	                                         byte mode)
	{
		FrameClass fc= frame1.getFrameClass();
		boolean hlc= FrameClass.lookup("higher_level_constraint").
		             subsumes(fc);
		_numberOfCorrespondences++;
		if (hlc) _numberOfHLCCorrespondences++;

		/////////////////////////////////////////////////////////////
		// Count correspondences for frame1's frame-class          //
		// (if frame1 and frame2 are of different frame-classes,   //
		//  only the first is counted, in order to make the counts //
		//  add up to the total number of correspondences).        //
		/////////////////////////////////////////////////////////////
		Integer count= (Integer)_numberOfCorrespondencesHashtable.
		                        get(fc);
		if (count == null) {
			count= new Integer(0);
		}
		count= new Integer(count.intValue() + 1);
		_numberOfCorrespondencesHashtable.put(fc, count);

		///////////////////////////////
		// Record the correspondence //
		///////////////////////////////

		if (mode == CorrespondenceHeuristics.PLAUSIBLE_CORRESPONDENCE)
		{
			Frame correspondence= new Frame("correspondence");
			correspondence.ssf("frame1", frame1);
			correspondence.ssf("frame2", frame2);
			//////////////////////////////////////////////
			// Assume equivalent until proved otherwise //
			//////////////////////////////////////////////
			correspondence.ssf("semantic_equivalence",
			  new DefinedTypeInstance(DefinedType.lookup
			                          ("semantic_equivalence"),
			  new EnumSymbol("equivalent")));
			frame1.aesf("corresponds_to", correspondence);
			frame2.aesf("corresponds_to", correspondence);
		} else {
			frame1.aesf("corresponds_to", frame2);
			frame2.aesf("corresponds_to", frame1);
		}

		///////////////////////////////////////////////////
		// If monogamy remembrance heuristic is enabled, //
		// remove HLCs from remembrance structures       //
		///////////////////////////////////////////////////

		if (hlc && CCUS._monogamy) {
			HLC_CCUS.forget1(frame1);
			HLC_CCUS.forget2(frame2);
		}

		///////////////////////////////
		// Report the correspondence //
		///////////////////////////////

		CCUS_FMA.indent();
		System.out.print(frame1);
		String f1d= ComparisonEngine._prettyPrint.
		            generateDescription(frame1, false);
		if (! hlc && ! f1d.equals(frame1.toString()))
			System.out.print(" (" + f1d + ")");
		if (mode == CorrespondenceHeuristics.PLAUSIBLE_CORRESPONDENCE)
			System.out.print(" plausibly");
		System.out.print(" corresponds to " + frame2);
		String f2d= ComparisonEngine._prettyPrint.
		            generateDescription(frame2, false);
		if (! hlc && ! f2d.equals(frame2.toString()))
			System.out.print(" (" + f2d + ")");
		System.out.println(".");

		//////////////////////////////////////////////////////
		// In CORRESPONDENCE mode,                          //
		// if the frames' "covers" slots are set,           //
		// record correspondence between the covered frames //
		//////////////////////////////////////////////////////

		if (mode == CorrespondenceHeuristics.CORRESPONDENCE) {
			Frame cover1= (Frame)frame1.gsf("covers");
			Frame cover2= (Frame)frame2.gsf("covers");
			if (cover1 != null && cover2 != null
			 && ! Correspondences.
			      correspondToEachOther(cover1,cover2))
			{
				Correspondences.recordCorrespondence
				                (cover1, cover2, mode);
			}
		}
	}

	/********************************************************************/
	// static Correspondences.retractCorrespondence :
	//                        (Frame x Frame x byte) -> void
	/*********************************************************************
	** Retracts a correspondence between two frames by deleting them from
	** each other's "corresponds_to" slot.
	** @param frame1 The first frame.
	** @param frame2 The second frame.
	*********************************************************************/

	public static void retractCorrespondence (Frame frame1, Frame frame2){
		FrameClass fc= frame1.getFrameClass();
		boolean hlc= FrameClass.lookup("higher_level_constraint").
		             subsumes(fc);
		_numberOfCorrespondences--;
		if (hlc) _numberOfHLCCorrespondences--;

		/////////////////////////////////////////////////////////////
		// Uncount correspondences for frame1's frame-class        //
		// (if frame1 and frame2 are of different frame-classes,   //
		//  only the first is counted, in order to make the counts //
		//  add up to the total number of correspondences).        //
		/////////////////////////////////////////////////////////////
		Integer count= (Integer)_numberOfCorrespondencesHashtable.
		                        get(fc);
		if (count == null) {
			System.err.println("Correspondences.retractCorrespond"
			+ "ence: _numberOfCorrespondencesHashtable.get("
			+ fc + ") == null");
			System.exit(1);
		}
		count= new Integer(count.intValue() - 1);
		if (count.intValue() == 0) {
			_numberOfCorrespondencesHashtable.remove(fc);
		} else {
			_numberOfCorrespondencesHashtable.put(fc, count);
		}

		//////////////////////////////
		// Erase the correspondence //
		//////////////////////////////

		Frame cf= Correspondences.correspondenceFrame(frame1,frame2);
		if (cf == null) {
			frame1.desf("corresponds_to", frame2);
			frame2.desf("corresponds_to", frame1);
		} else {
			frame1.desf("corresponds_to", cf);
			frame2.desf("corresponds_to", cf);
			Frame.remove(cf);
		}

		Vector setOfFrames= (Vector)_alreadyComparedAndFailed.
		                            get(frame1);
		if (setOfFrames == null) {
			setOfFrames= new Vector();
			_alreadyComparedAndFailed.put(frame1, setOfFrames);
		}
		setOfFrames.addElement(frame2);

		/////////////////////////////////////////
		// Report the retracted correspondence //
		/////////////////////////////////////////

		CCUS_FMA.indent();
		System.out.println("Correspondence between " + frame1
		+ " and " + frame2 + " retracted.");
	}

	/********************************************************************/
	// static Correspondences.correspondToEachOther :
	//                        (Frame x Frame) -> boolean
	/*********************************************************************
	** Returns true iff two given frames correspond to each other
	** (i.e. iff they are in each other's "corresponds_to" slots,
	** or if they have a reified correspondence frame between them).
	** @param frame1 The first frame.
	** @param frame2 The second frame.
	** @return Whether they correspond.
	*********************************************************************/

	public static boolean correspondToEachOther (Frame frame1,
	                                             Frame frame2)
	{
		Vector frame1c= (Vector)frame1.gsf("corresponds_to");
		Vector frame2c= (Vector)frame2.gsf("corresponds_to");

		return frame1c != null && frame2c != null
		    && (frame1c.contains(frame2) && frame2c.contains(frame1)
		        || correspondenceFrame(frame1,frame2) != null);
	}

	/********************************************************************/
	// static Correspondences.correspondenceFrame :
	//                        (Frame x Frame) -> Frame
	/*********************************************************************
	** If two given frames correspond to each other then returns the
	** correspondence frame if one exists, otherwise returns null.
	** @param frame1 The first frame.
	** @param frame2 The second frame.
	** @return The correspondence frame between them if there is one,
	**         otherwise null.
	*********************************************************************/

	public static Frame correspondenceFrame (Frame frame1, Frame frame2) {
		Vector frame1c= (Vector)frame1.gsf("corresponds_to");
		if (frame1c == null) return null;
		Vector frame2c= (Vector)frame2.gsf("corresponds_to");
		if (frame2c == null) return null;

		ListIterator li1= frame1c.listIterator();
		while (li1.hasNext()) {
			Frame cf= (Frame)li1.next();
			if (FrameClass.lookup("correspondence").
			    subsumes(cf.getFrameClass()))
			{
				ListIterator li2= frame2c.listIterator();
				while (li2.hasNext()) {
					if (li2.next() == cf) {
						return cf;
					}
				}
			}
		}
		return null;
	}

	/********************************************************************/
	// static Correspondences.alreadyComparedAndFailed :
	//                        (Frame x Frame) -> boolean
	/*********************************************************************
	** Returns true iff two given frames have already been compared and
	** failed to correspond.
	** @param frame1 The first frame.
	** @param frame2 The second frame.
	** @return Whether they have already been compared and failed to
	**         correspond.
	*********************************************************************/

	public static boolean alreadyComparedAndFailed (Frame frame1,
	                                                Frame frame2)
	{
		Vector setOfFrames= (Vector)_alreadyComparedAndFailed.
		                            get(frame1);
		if (setOfFrames == null) return false;
		ListIterator li= setOfFrames.listIterator();
		while (li.hasNext()) {
			if (li.next() == frame2) return true;
		}
		return false;
	}
}
