      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // DefinedType.java (2000:10:11 12:00) //////
//////////////////////////////////////////////
// Methods of DefinedType:             ///////
// - DefinedType constructor           ///////
// - getTypeFrame                      ///////
// - getName                           ///////
// - toString                          ///////
// - static createDefinedTypes         ///////
// - static getAllFrameClasses         //////
// - static lookup(String)             /////
// - static lookup(String,Schema)      ////
// - static writeClass                 ///
// - static readClass                  //
////////////////////////////////////////

import java.io.*;
import java.util.*;

import jex.*;

/*****************************************************************************
** The class `DefinedType' represents a defined type which can have instances
** in the frame system, doing for <A HREF="DefinedTypeInstance.html">
** DefinedTypeInstance</A> what <A HREF="FrameClass.html">FrameClass</A> does
** for <A HREF="Frame.html">Frame</A>.
*****************************************************************************/

public class DefinedType implements Serializable {
	private Type _underlyingType;
	private String _name;
	private static Hashtable/*String,DefinedType*/ _definedTypeIndex
	                                               = new Hashtable();

	/********************************************************************/
	// DefinedType constructor : Type ->
	/*********************************************************************
	** Constructs this defined type by plugging in the given Jex Type
	** object which is taken to represent it (it does so by copying the
	** reference, which is maximally efficient, but the client should not
	** then alter the Type directly, as that would compromise the
	** integrity of the frame).
	** @param definedTypeType A Jex Type object to wrap in this frame.
	*********************************************************************/

	public DefinedType (Type definedTypeType) {
		_underlyingType= definedTypeType;
		try {
			_name= definedTypeType.getId();
		}
		catch (UnsetValueException uve) {
			System.err.println("UnsetValueException caught in "
			+ "DefinedType constructor");
			System.exit(1);
		}
		if (_definedTypeIndex.get(_name) != null) {
			System.err.println("DefinedType \"" + _name + "\" "
			+ "already exists!");
			System.exit(1);
		}

		_definedTypeIndex.put(_name,this);
	}

	/********************************************************************/
	// DefinedType.getTypeFrame : void -> Frame
	/*********************************************************************
	** Delivers up the underlying frame of this defined type.
	** @return The class frame representing this defined type.
	*********************************************************************/
 
	public Frame getTypeFrame () {
		return JexToFrames.getCachedFrame(_underlyingType);
	}
	
	/********************************************************************/
	// DefinedType.getName : void -> String
	/*********************************************************************
	** Returns the name of this defined type.
	** @return The name of this defined type.
	*********************************************************************/
 	
	public String getName () {
		return _name;
	}
	
	/********************************************************************/
	// DefinedType.toString : void -> String
	/*********************************************************************
	** Returns a string representation of this defined type.
	** @return The string representation of this defined type.
	*********************************************************************/

	public String toString () {
		return _name;
	}

	/********************************************************************/
	// static DefinedType.createDefinedTypes : Schema -> void
	/*********************************************************************
	** Create a DefinedType for every defined type in a given schema.
	** @param schema The schema.
	*********************************************************************/

	public static void createDefinedTypes (Schema schema) {
		ListIterator definedTypes= schema.getTypes();
		while (definedTypes.hasNext()) {
			Type type= (Type)definedTypes.next();
			new DefinedType(type);
		}
	}

	/********************************************************************/
	// static DefinedType.getAllDefinedTypes : void -> Enumeration
	/*********************************************************************
	** Returns all the defined types in the frame system.
	** @return An enumeration of all the defined types in the frame
	**         system.
	*********************************************************************/

	public static Enumeration getAllDefinedTypes () {
		return _definedTypeIndex.elements();
	}

	/********************************************************************/
	// static DefinedType.lookup : String -> DefinedType
	/*********************************************************************
	** Looks up the defined type with the given name. A quick and
	** convenient method which assumes that the defined type of that name
	** (already) exists - if not, it will report an error and exit from
	** within.
	** @param definedTypeName The name of the frame class to return.
	** @return The DefinedType object whose name is `definedTypeName'.
	** @see lookup(String,Schema)
	*********************************************************************/
 
	public static DefinedType lookup (String definedTypeName) {
		DefinedType definedType= (DefinedType)_definedTypeIndex.
			get(definedTypeName);
		if (definedType == null) {
			System.err.println("Defined type \"" + definedTypeName
			+ "\" does not exist [tip: use DefinedType.lookup"
			+ "(String,Schema) for greater flexibility than "
			+ "DefinedType.lookup(String)]");
			System.exit(1);
		}
		return definedType;
	}

	/********************************************************************/
	// static DefinedType.lookup : String x Schema -> DefinedType
	/*********************************************************************
	** Looks up the defined type with the given name. A more cumbersome
	** version of the lookup method, for use when the defined type of that
	** name may not have been constructed yet. If the defined type does
	** not exist, it tries to construct it, looking up the defined type of
	** that name in the given EXPRESS schema. Failing that, it throws a
	** DefinedTypeDoesNotExist exception.
	** @param definedTypeName The name of the defined type to return.
	** @param frameClassModel The EXPRESS model in which `definedTypeName'
	**  is defined, in case that defined type hasn't been constructed yet.
	** @return The DefinedType object whose name is `definedTypeName'.
	** @exception DefinedTypeDoesNotExist If no defined type of that name
	**  exists already, and there is no defined type of that name in the
	**  frame class schema either.
	** @see lookup(String)
	*********************************************************************/
  
	public static DefinedType lookup (String definedTypeName,
	                                  Schema frameClassModel)
	throws DefinedTypeDoesNotExist
	{
		DefinedType definedType= (DefinedType)_definedTypeIndex.
			get(definedTypeName);
		if (definedType == null) {
			Type t= null;
			try {
				t= frameClassModel.findType(definedTypeName);
			}
			catch (NullParameterException npe) {
				System.err.println("NullParameterException ca"
				+"ught in DefinedType.lookup(String,Schema)");
				System.exit(1);
			}
			if (t != null)
				definedType= new DefinedType(t);
			else
				throw new DefinedTypeDoesNotExist();
		}
		return definedType;
	}

	/********************************************************************/
	// static DefinedType.writeClass : ObjectOutput -> void
	/*********************************************************************
	** Serialises the static data members (class variables) of
	** DefinedType.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void writeClass (ObjectOutput oo) throws IOException {
		oo.writeObject(_definedTypeIndex);
	}

	/********************************************************************/
	// static DefinedType.readClass : ObjectInputStream -> void
	/*********************************************************************
	** Deserialises the static data members (class variables) of
	** DefinedType.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void readClass (ObjectInput oi) throws IOException {
		try {
			_definedTypeIndex= (Hashtable)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the st"
			+ "upid compiler is just too dumb to realise this!");
		}
	}
}
