      /////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM  //
    //  (Version 2)                         ///
   // Written by Andrew Broad              ////
  //                                      /////
 // ExpressCUS.java (2001:09:20 15:29)   //////
///////////////////////////////////////////////
// Methods of ExpressCUS:               ///////
// - static getExtractionCaseLibrary    ///////
// - static getNavigationCaseLibrary    ///////
// - static getPrettyPrintCaseLibrary   ///////
// - static getSubexpressionCaseLibrary ///////
// - static getSemanticCaseLibrary      ///////
// - static writeClass                  ///////
// - static readClass                   ///////
// - static main                        //////
// - static usage                       /////
// - static processArguments            ////
// - static processFlags                ///
// - time                               //
/////////////////////////////////////////

import java.util.*;
import java.io.*;

import jex.*;

/*****************************************************************************
** The class `ExpressCUS' is the top-level class of the EXPRESS Constraint-
** Understanding System. It is not meant to be instantiated.
*****************************************************************************/

public class ExpressCUS {
	public static boolean _generateHTML= false;
	public static boolean _onlyInputModelFramesInHTML= false;
	public static boolean _showTimes= false;

	private static CaseLibrary _extractionCaseLibrary= new CaseLibrary();
	private static CaseLibrary _navigationCaseLibrary= new CaseLibrary();
	private static CaseLibrary _ppCaseLibrary= new CaseLibrary();
	private static CaseLibrary _subexprCaseLibrary= new CaseLibrary();
	private static CaseLibrary _semanticCaseLibrary= new CaseLibrary();

	private static String _expressFilename= null;
	private static String _caseLibraryFilename
	       = "/home/broada/jws/PhD_CUS/CASES/library.cas";
	private static boolean _parseFirst= false;

	/********************************************************************/
	// static ExpressCUS.getExtractionCaseLibrary : void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the extraction case library of the EXPRESS
	** CUS.
	** @return The extraction case library.
	*********************************************************************/

	public static CaseLibrary getExtractionCaseLibrary () {
		return _extractionCaseLibrary;
	}

	/********************************************************************/
	// static ExpressCUS.getNavigationCaseLibrary : void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the navigation case library of the EXPRESS
	** CUS.
	** @return The navigation case library.
	*********************************************************************/

	public static CaseLibrary getNavigationCaseLibrary () {
		return _navigationCaseLibrary;
	}

	/********************************************************************/
	// static ExpressCUS.getPrettyPrintCaseLibrary : void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the pretty-print case library of the EXPRESS
	** CUS.
	** @return The pretty print case library.
	*********************************************************************/

	public static CaseLibrary getPrettyPrintCaseLibrary () {
		return _ppCaseLibrary;
	}

	/********************************************************************/
	// static ExpressCUS.getSubexpressionCaseLibrary : void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the subexpression case library of the
	** EXPRESS CUS.
	** @return The subexpression case library.
	*********************************************************************/

	public static CaseLibrary getSubexpressionCaseLibrary () {
		return _subexprCaseLibrary;
	}

	/********************************************************************/
	// static ExpressCUS.getSemanticCaseLibrary : void -> CaseLibrary
	/*********************************************************************
	** Returns a reference to the semantic-equivalence/strength case
	** library of the EXPRESS CUS.
	** @return The semantic-equivalence/strength case library.
	*********************************************************************/

	public static CaseLibrary getSemanticCaseLibrary () {
		return _semanticCaseLibrary;
	}

	/********************************************************************/
	// static ExpressCUS.writeClass : ObjectOutput -> void
	/*********************************************************************
	** Serialises the static data members (class variables) of ExpressCUS.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_extractionCaseLibrary);
		oo.writeObject(_navigationCaseLibrary);
		oo.writeObject(_ppCaseLibrary);
		oo.writeObject(_subexprCaseLibrary);
		oo.writeObject(_semanticCaseLibrary);
	}

	/********************************************************************/
	// static ExpressCUS.readClass : ObjectInputStream -> void
	/*********************************************************************
	** Deserialises the static data members (class variables) of
	** ExpressCUS.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_extractionCaseLibrary= (CaseLibrary)oi.readObject();
			_navigationCaseLibrary= (CaseLibrary)oi.readObject();
			_ppCaseLibrary= (CaseLibrary)oi.readObject();
			_subexprCaseLibrary= (CaseLibrary)oi.readObject();
			_semanticCaseLibrary= (CaseLibrary)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the st"
			+ "upid compiler is just too dumb to realise this!");
		}
	}

	/********************************************************************/
	// static ExpressCUS.main : String -> void
	/*********************************************************************
	** The main entry point to the EXPRESS Constraint-Understanding
	** System.
	** @param argv The vector of command line arguments. There should be
	**  two of these: the filename of an EXPRESS model (.exp) to
	**  parse using jex, and a case library (.cas). If the latter is not
	**  specified, it defaults to "CASES/library.cas". There can also be
	**  command-line flags:
	**  -p causes the HLCs in the summary to be pretty-printed,
	**  -g causes HTML files for frames to be generated without asking,
	**  -i causes only input model frames to be included in the HTML,
	**  -s causes the intermediate output of the CUS to be suppressed,
	**  -t prints timestamps at the major phase boundaries.
	*********************************************************************/

	public static void main (String argv[]) {
		int numberOfFrames= -1;
		try {
			System.setProperty("java.compiler","none");
			System.setProperty("jex.licenseLocation",
			                   "27000@manatee");

			processArguments(argv);

			Frame._markFramesAsInputModel= true;

			Model inputModel= null;

			if (_parseFirst) {
				System.out.println(time() + "Parsing EXPRESS "
				+ "model in \"" +_expressFilename + "\" using"
				+ " Jex...");
				inputModel= (Model)Model.readText
				            (new File(_expressFilename));
				System.out.println
				           ("Done parsing EXPRESS model.\n");

				System.out.println(time() + "Loading case lib"
				+"rary from \""+_caseLibraryFilename+"\"...");
				CaseLibrary.load(_caseLibraryFilename);
				System.out.println
				 ("Done loading case library.\n");
			} else {
				System.out.println(time() + "Loading case lib"
				+"rary from \""+_caseLibraryFilename+"\"...");
				CaseLibrary.load(_caseLibraryFilename);
				System.out.println
				           ("Done loading case library.\n");

				System.out.println(time() + "Parsing EXPRESS "
				+ "model in \"" +_expressFilename + "\" using"
				+ " Jex...");
				inputModel= (Model)Model.readText
				            (new File (_expressFilename));
				System.out.println
				 ("Done parsing EXPRESS model.\n");
			}

			System.out.println(time() + "Converting Jex objects "
			+ "to frames...");
			Frame._numberOfFrames= 0;
	 		Frame inputModelFrameNetwork= JexToFrames.convert
	 		                              (inputModel,false);
			numberOfFrames= Frame._numberOfFrames;
	 		System.out.println("Done converting Jex objects to "
	 		                 + "frames.\n");

			System.out.println(time() + "Extracting higher-level "
			+ "knowledge...\n");
			ExtractionEngine.extraction(inputModelFrameNetwork,
			                            _extractionCaseLibrary,
			                            _navigationCaseLibrary,
			                            _ppCaseLibrary);
		} catch (Throwable t) {
			System.err.println(
				"----------------------------------------" +
				"----------------------------------------");
			System.err.println(time() + "A Throwable of type "
			+ t.getClass() + " has been caught in "
			+ "ExpressCUS.main!\n\ttoString == " + t + "\n");
			//System.err.println("Its Message is:-\n"
			//                  + t.getMessage() + "\n");
			//System.err.println("Its LocalizedMessage is:-\n"
			//+ t.getLocalizedMessage());
			if (! (t instanceof Error)) {
				System.out.print(time() + "Print stack trace?"
				+ " (y/n)? ");
				BufferedReader input= new BufferedReader
				           (new InputStreamReader(System.in));
				String printStackTrace= null;
				try {
					printStackTrace= input.readLine();
				} catch (IOException ioe) {
					System.err.println("ExpressCUS.main: "
					+ "IOException thrown by BufferedRead"
					+ "er.readLine()");
					System.exit(1);
				}
				System.out.println();
				if (printStackTrace.
				    regionMatches(true,0,"y",0,1))
				{
					t.printStackTrace();
				}
			} else {
				System.err.println("Not printing stack trace "
				+ "because " + t + " instanceof Error");
			}
		}
		System.out.println(
			"----------------------------------------" +
			"----------------------------------------");
		///////////////////////////////////
		// Print experimental statistics //
		///////////////////////////////////
		System.out.println("Number of frames in model: "
		+ numberOfFrames);
		System.out.println("Number of HLCs extracted: "
		+ ExtractionEngine._numberOfHLCsExtracted);
		System.out.println(
			"----------------------------------------" +
			"----------------------------------------");
		if (! _generateHTML) {
			System.out.print(time() + "Generate HTML for the fram"
			+ "es (y/n)? ");
			BufferedReader input= new BufferedReader
			                   (new InputStreamReader(System.in));
			String generateHTML= null;
			try {
				generateHTML= input.readLine();
			} catch (IOException ioe) {
				System.err.println("ExpressCUS.main: IOExcept"
				+ "ion thrown by BufferedReader.readLine()");
				System.exit(1);
			}
			System.out.println();
			_generateHTML= generateHTML.
			               regionMatches(true,0,"y",0,1);
		}
		if (_generateHTML) {
			System.out.println("\n" + time() + "Generating HTML "
			+ "in /usr/local/tmp/users/broada/FRAMES/...");
			Frame.generateFIL(true, _onlyInputModelFramesInHTML,
			                  null);
		}
		if (_showTimes) {
			System.out.println("\n" + time() + "Finished.");
		}
	}

	/********************************************************************/
	// static ExpressCUS.usage : void -> void
	/*********************************************************************
	** Print the usage report when the command-line arguments are invalid,
	** and exit the system.
	*********************************************************************/

	private static void usage () {
		System.err.println("Usage: Two arguments: an EXPRESS model (."
		+ "exp file) and a case library (.cas file). The latter is op"
		+"tional - if omitted, the default is \"CASES/library.cas\"");
		System.err.println("    Flags:-");
		System.err.println("-p: Pretty-print HLCs in summary");
		System.err.println("-g: Generate HTML without asking");
		System.err.println("-i: Include only input model frames in "
		+ "HTML");
		System.err.println("-s: Suppress intermediate output");
		System.err.println("-t: Show times at phase boundaries");
		System.exit(1);
	}

	/********************************************************************/
	// static ExpressCUS.processArguments : String[] -> void
	/*********************************************************************
	** A subroutine to process the command-line arguments to the CCUS.
	** @param argv The vector of command-line arguments. There should be
	**  three of these: the filenames of two EXPRESS models (.exp) to
	**  parse using jex, and a case library (.cas). If the latter is not
	**  specified, it defaults to "CASES/library.cas".
	*********************************************************************/

	private static void processArguments (String argv[]) {
		for (int i= 0; i < argv.length; i++) {
			if (argv[i].endsWith(".exp")) {
				_expressFilename= argv[i];
				_parseFirst= false;
			}
			if (argv[i].endsWith(".cas")) {
				_caseLibraryFilename= argv[i];
				_parseFirst= true;
			}
			if (argv[i].startsWith("-")) {
				processFlags(argv[i]);
			}
		}

		if (_expressFilename == null) {
			usage();
		}
	}

	/********************************************************************/
	// static ExpressCUS.processFlags : String[] -> void
	/*********************************************************************
	** A subroutine to process the flags in a command-line argument
	** starting with `-'. This method assumes that it is only called with
	** strings starting with `-', and ignores the first character.
	** @param arg A `flags' command-line argument (i.e. `-' followed by
	**            letters).
	*********************************************************************/

	private static void processFlags (String arg) {
		char[] flags= arg.toCharArray();

		for (int i= 1; i < flags.length; i++) {
			switch (flags[i]) {
				case 'p':
					ExtractionEngine.
					_prettyPrintHLCs= true;
					break;
				case 's':
					ExtractionEngine.
					_showIntermediateOutput= false;
					break;
				case 'g':
					_generateHTML= true;
					break;
				case 'i':
					_onlyInputModelFramesInHTML= true;
					break;
				case 't':
					_showTimes= true;
					break;
				default:
					System.err.println("Warning: -"
					+ flags[i] + " not recognised.");
			}
		}
	}

	/********************************************************************/
	// static ExpressCUS.time : void -> String
	/*********************************************************************
	** If _showTimes is true, return the current time as a string of the
	** form "(HH:MM:SS.SSS) ", otherwise return the empty string.
	** @return The time string.
	*********************************************************************/

	private static String time () {
		String time= "";
		if (_showTimes) {
			GregorianCalendar gregorc= new GregorianCalendar();
			//////////////
			// hours... //
			//////////////
			time= "(";
			int hours= gregorc.get(GregorianCalendar.HOUR_OF_DAY);
			if (hours < 10) time+= "0";
			time+= hours + ":";
			////////////////
			// minutes... //
			////////////////
			int minutes= gregorc.get(GregorianCalendar.MINUTE);
			if (minutes < 10) time+= "0";
			time+= minutes + ":";
			////////////////
			// seconds... //
			////////////////
			int seconds= gregorc.get(GregorianCalendar.SECOND);
			if (seconds < 10) time+= "0";
			time+= seconds + ".";
			/////////////////////
			// milliseconds... //
			/////////////////////
			int ms= gregorc.get(GregorianCalendar.MILLISECOND);
			if (ms < 100) {
				time+= "0";
				if (ms < 10) time+= "0";
			}
			time+= ms + ") ";
		}
		return time;
	}
}
