      //////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM   //
    //  (Version 2)                          ///
   // Written by Andrew Broad               ////
  //                                       /////
 // FrameTemplate.java (2001:03:05 14:23) //////
////////////////////////////////////////////////
// Methods of FrameTemplate:              //////
// - static find                          //////
// - FrameTemplate constructor            //////
// - getFrameClass                        //////
// - getSlotFiller(Slot)                  //////
// - getSlotFiller(String)                //////
// - setSlotFiller(Slot,Object)           //////
// - setSlotFiller(String,Object)         //////
// - getSlots                             //////
// - toString                             /////
// - print                                ////
// - static writeClass                    ///
// - static readClass                     //
///////////////////////////////////////////

import java.io.*;
import java.util.*;

import jex.*;

/*****************************************************************************
** The class `FrameTemplate' represents a frame template, which encodes how to
** convert a given Java (in particular, Jex) object to a frame.
*****************************************************************************/

public class FrameTemplate implements FrameInterface, Serializable {
	private Hashtable/*Slot->Object*/ _frameRep= new Hashtable();
	private Vector/*Slot*/ _slots= new Vector();
	private FrameClass _frameClass;
	private Class _jClass;
	private String _jId;
	private static Hashtable/*Class->FrameTemplate*/ _allFrameTemplates
	                                                 = new Hashtable();

	/********************************************************************/
	// static FrameTemplate find : Class -> FrameTemplate
	/*********************************************************************
	** Finds a frame template that is appropriate to a given Java (Jex)
	** object class.
	** @param jClass The Java (Jex) class that you want to find a frame
	**               template to convert.
	** @return A FrameTemplate for converting objects of type jClass to
	**         frames, or null if no suitable frame template exists.
	*********************************************************************/

	public static FrameTemplate find (Class jClass) {
		return (FrameTemplate)_allFrameTemplates.get(jClass);
	}

	/********************************************************************/
	// FrameTemplate constructor : FrameClass x Class x String ->
	/*********************************************************************
	** Constructs this frame template.
	** @param frameClass The frame class of the frame template to be
	**                   constructed, which specifies the set of slots the
	**                   frame template should have.
	** @param jClass The Java class associated with this frame template.
	** @param jId The name of the Java object identifier associated with
	**            this frame template.
	*********************************************************************/
 
 	public FrameTemplate (FrameClass frameClass, Class jClass, String jId)
 	{
		_frameClass= frameClass;
		_jClass= jClass;
		if (_allFrameTemplates.get(jClass) != null) {
			System.err.println("A FrameTemplate already exists "
			+ "for converting objects of type " + jClass.getName()
			+ "!");
			System.exit(1);
		}
		_allFrameTemplates.put(jClass,this);
		_jId= jId;
	}
	
	/********************************************************************/
	// FrameTemplate.getFrameClass : void -> FrameClass
	/*********************************************************************
	** Returns the frame class of this frame template.
	** @return The frame class of this frame template.
	*********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}

	/********************************************************************/
	// FrameTemplate.getSlotFiller : Slot -> Object
	/*********************************************************************
	** Returns the filler of a given slot.
	** @param slot The slot whose filler you want to read.
	** @return The filler of `slot', as held in this frame.
	*********************************************************************/
 
 	public Object getSlotFiller (Slot slot) {
		return _frameRep.get(slot);
	}

	/********************************************************************/
	// FrameTemplate.getSlotFiller : String -> Object
	/*********************************************************************
	** Returns the filler of a given slot.
	** @param slot The name of the slot whose filler you want to read.
	** @return The filler of the slot, as held in this frame.
	** @exception InvalidSlot If the frame templatedoes not have a slot of
	**                        that name (as specified in its frame class).
	*********************************************************************/

	public Object getSlotFiller (String slotName)
	                            throws InvalidSlot
	{
		return getSlotFiller(getFrameClass().getSlot(slotName));
	}

	/********************************************************************/
	// FrameTemplate.setSlotFiller : (Slot x Object) -> void
	/*********************************************************************
	** Fills a given slot with a given value. Records this slot as having
	** been written directly.
	** @param slot The slot you want to write into.
	** @param filler The filler (value) you want to write into `slot',
	**               which must be of the right type for `slot'.
	*********************************************************************/
 
	public void setSlotFiller (Slot slot, Object filler) {
		if (filler != null) {
			_frameRep.put(slot,filler);
			if (! _slots.contains(slot)) {
				_slots.add(slot);
			}
		} else {
			_frameRep.remove(slot);
			_slots.remove(slot);
		}
	}

	/********************************************************************/
	// FrameTemplate.setSlotFiller : (String x Object) -> void
	/*********************************************************************
	** Fills a given slot with a given value.
	** @param slot The name of the slot you want to write into.
	** @param filler The filler (value) you want to write into the slot,
	**               which must be of the right type for the slot.
	** @exception InvalidSlot If the frame template does not have that
	**                        slot (as specified in its frame class).
	*********************************************************************/
 
	public void setSlotFiller (String slotName, Object filler)
	                          throws InvalidSlot
	{
		setSlotFiller(getFrameClass().getSlot(slotName),filler);
	}

	/********************************************************************/
	// FrameTemplate.getSlots : void -> ListIterator
	/*********************************************************************
	** Returns an iterator over the filled slots in this frame template,
	** in the order that they were filled. (If a slot was filled twice,
	** its position in the iterator is determined by the first time it was
	** filled.)
	** @return An iterator over the filled slots in this frame template.
	*********************************************************************/

	public ListIterator getSlots () {
		return _slots.listIterator();
	}

	/********************************************************************/
	// FrameTemplate.toString : void -> String
	/*********************************************************************
	** Returns a string representing this frame template. The format is
	** <tt><i>frame-class</i> (<i>Java-class</i> <i>identifier</i>)</tt>.
	** @return The string representation of this frame template.
	*********************************************************************/

	public String toString () {
		return getFrameClass() +
		       " (" + _jClass.getName() + _jId + ")";
	}

	/********************************************************************/
	// FrameTemplate.print : PrintStream -> void
	/*********************************************************************
	** Prints the frame template in human-readable text form. Any
	** references are just printed as they are.
	** @param ps A stream to print the frame template to.
	*********************************************************************/

	public void print (PrintStream ps) {
		Slot slot;

		//generateFILJ(new PrintWriter(ps,true),false);
			System.err.println
			("FrameTemplate.generateFILJ not implemented");
			System.exit(1);
		ps.println();
	}

	/********************************************************************/
	// static FrameTemplate.writeClass : ObjectOutput -> void
	/*********************************************************************
	** Serialises the static data members (class variables) of
	** FrameTemplate.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_allFrameTemplates);
	}

	/********************************************************************/
	// static FrameTemplate.readClass : ObjectInputStream -> void
	/*********************************************************************
	** Deserialises the static data members (class variables) of
	** FrameTemplate.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_allFrameTemplates= (Hashtable)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the st"
			+ "upid compiler is just too dumb to realise this!");
		}
	}
}
