     //////////////////////////////////////////////
    // THE COMPARATIVE CODE-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad                   ///
  //                                           ////
 // SemanticStrength.java (2002:01:14 18:52)  /////
///////////////////////////////////////////////////
// Methods of SemanticStrength:              //////
// - useSemanticCase                         /////
// - compareSSSfillers                       ////
// - compareDoubles                          ///
// - logicalToDouble                         //
//////////////////////////////////////////////

import java.io.*;
import java.util.*;
import jex.*;

/*****************************************************************************
** The class `SemanticStrength' implements the methods pertaining to the use
** of semantic equivalence/strength cases.
** This class is not meant to be instantiated.
*****************************************************************************/

public class SemanticStrength {
	public final static int GREATER= 1;
	public final static int LESS= -1;
	public final static int EQUAL= 0;
	public final static int INCOMPARABLE= 2;

	/********************************************************************/
	// SemanticStrength.useSemanticCase : (Frame x Frame x Slot x Case)
	//                                    -> Case
	/*********************************************************************
	** This method, which assumes that the CCUS is currently matching in
	** PLAUSIBLE_CORRESPONDENCE mode, uses a semantic-equivalence/strength
	** case to decide whether to fail the match, or which frame is
	** stronger. It takes as input a semantic case, which may be null -
	** if so, it attempts to retrieve <B>one</B> semantic case to use
	** (some situations will require multiple semantic cases, but they are
	** not supported yet).
	** @param frame1 The frame to match in the first model.
	** @param frame2 The frame to match in the second model.
	** @param slot The slot for which a difference has been found.
	** @param seCase The semantic-equivalence/strength case that pertains
	**               to the current frame-match, if there is one.
	** @return seCase (if a semantic case is retrieved, the caller
	**         matchFrames needs a reference to it).
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @exception FailMatch Tells the caller matchFrames to fail the match
	**                      if the difference means that the frames are
	**                      semantically incomparable. It's the caller's
	**                      responsibility to retract the correspondence.
	*********************************************************************/

	public static Case useSemanticCase (Frame frame1, Frame frame2,
	                                    Slot slot, Case seCase,
	                                    EntityPair entityPair)
	                                   throws FailMatch
	{	
		/////////////////////////////////////////////
		// Use semantic-equivalence/strength cases //
		// in PLAUSIBLE_CORRESPONDENCE mode        //
		/////////////////////////////////////////////
		if (seCase == null)
		{
			seCase= ExpressCUS.getSemanticCaseLibrary().
			                   retrieveOneMatchingCase(frame1);
			/////////////////////////////////////////////////////
			// The match fails immediately if no semantic case //
			// was retrieved, or if the second frame doesn't   //
			// match the case too (we know the first frame     //
			// matches the case from retrieveOneMatchingCase)  //
			/////////////////////////////////////////////////////
			if (seCase == null || ! seCase.match(frame2)) {
				throw new FailMatch();
			}
		}
		// {seCase != null}
		Frame solution= (Frame)seCase.getCaseFrame().gsf("solution");

		////////////////////////////////////////////////////////////
		// Necessary slots must match - otherwise, fail the match //
		// and retract the correspondence because the frames are  //
		// semantically incomparable.                             //
		////////////////////////////////////////////////////////////
		Vector necessarySlots=(Vector)solution.gsf("necessary_slots");
		ListIterator ns= necessarySlots.listIterator();
		while (ns.hasNext()) {
			if (((FILString)ns.next()).toString().
			    equals(slot.getName()))
			{
				throw new FailMatch();
			}
		}

		//////////////////////////////////////////////////////////////
		// Semantic-strength slots affect which frame is stronger - //
		// either positively (the stronger frame has the greater    //
		// value for the slot) or negatively.                       //
		//////////////////////////////////////////////////////////////
		Vector semanticStrengthSlots
		= (Vector)solution.gsf("semantic_strength_slots");
		Frame correspondenceFrame= null;
		if (semanticStrengthSlots.size() > 0) {
			correspondenceFrame
			= Correspondences.correspondenceFrame(frame1, frame2);
		}
		ListIterator sss= semanticStrengthSlots.listIterator();
		while (sss.hasNext()) {
			Frame sssFrame= (Frame)sss.next();
			if (((FILString)sssFrame.gsf("slot_name")).toString().
			    equals(slot.getName()))
			{
				String f1d= ComparisonEngine._prettyPrint.
				            generateDescription(frame1,false);
				if (! f1d.equals(frame1.toString()))
					f1d= frame1 + (" (" + f1d + ")");
				String f2d= ComparisonEngine._prettyPrint.
				            generateDescription(frame2,false);
				if (! f2d.equals(frame2.toString()))
					f2d= frame2 + (" (" + f2d + ")");

				CCUS_FMA.indent();
				int first= compareSSSfillers
				           (frame1.getSlotFiller(slot),
				            frame2.getSlotFiller(slot),
				            entityPair);
				if (sssFrame.gsf("influence").toString().
				    equals("positive"))
				{
					if (first == GREATER) {
						System.out.println(f1d + " is"
						+" semantically stronger than"
						+" " + f2d + ".");

						correspondenceFrame.ssf
						("semantic_equivalence", new
						 DefinedTypeInstance
						 (DefinedType.lookup
						  ("semantic_equivalence"),
						 new EnumSymbol
						 ("frame1_stronger")));
					} else if (first == LESS) {
						System.out.println(f1d + " is"
						+" semantically weaker than "
						+ f2d + ".");

						correspondenceFrame.ssf
						("semantic_equivalence", new
						 DefinedTypeInstance
						 (DefinedType.lookup
						  ("semantic_equivalence"),
						 new EnumSymbol
						 ("frame2_stronger")));
					}
				}
				else //negative
				{
					if (first == LESS) {
						System.out.println(f1d + " is"
						+" semantically stronger than"
						+" " + f2d + ".");

						correspondenceFrame.ssf
						("semantic_equivalence", new
						 DefinedTypeInstance
						 (DefinedType.lookup
						  ("semantic_equivalence"),
						 new EnumSymbol
						 ("frame1_stronger")));
					} else if (first == GREATER){
						System.out.println(f1d + " is"
						+" semantically weaker than "
						+ f2d + ".");

						correspondenceFrame.ssf
						("semantic_equivalence", new
						 DefinedTypeInstance
						 (DefinedType.lookup
						  ("semantic_equivalence"),
						 new EnumSymbol
						 ("frame2_stronger")));
					}
				}
				if (first == EQUAL) {
					System.out.println(f1d +" is "
					+ "semantically equivalent to "
					+ f2d + ".");

					//correspondenceFrame.ssf/////////////
					//("semantic_equivalence", new////////
					///DefinedTypeInstance(DefinedType.///
					////lookup("semantic_equivalence"),///
					///new EnumSymbol("equivalent")));////
				} else if (first == INCOMPARABLE) {
					System.out.println(f1d +" is "
					+ "semantically incomparable to "
					+ f2d + ".");

					correspondenceFrame.ssf
					("semantic_equivalence", new
					 DefinedTypeInstance(DefinedType.
					  lookup("semantic_equivalence"),
					 new EnumSymbol("incomparable")));

					throw new FailMatch();
				}
			}
		}
		return seCase;
	}

	/********************************************************************/
	// SemanticStrength.compareSSSfillers : (Object x Object) -> int
	/*********************************************************************
	** Compares two semantic-strength-slot-fillers, whatever their types,
	** and returns GREATER if the first is 'greater than' the second, LESS
	** if the first is 'less than' the second, EQUAL if they are equal
	** (semantically equivalent), or INCOMPARABLE if they are semantically
	** incomparable.
	** @param filler1 The first semantic-strength-slot-filler.
	** @param filler2 The second semantic-strength-slot-filler.
	** @param entityPair The current entity-pair for the summarised output
	**                   (optional).
	** @return GREATER, LESS, EQUAL or INCOMPARABLE.
	*********************************************************************/

	private static int compareSSSfillers (Object filler1, Object filler2,
	                                      EntityPair entityPair)
	{
		/////////////////////////////////////////////////////////////
		// _literal frames: return the literal value - as a double //
		/////////////////////////////////////////////////////////////
		if (filler1 instanceof Frame && filler2 instanceof Frame) {
			double a= ((Number)((Frame)filler1).
			          gsf("literal_value")).doubleValue();
			double b= ((Number)((Frame)filler2).
			          gsf("literal_value")).doubleValue();
			return compareDoubles(a,b);
		}
		//////////////////////////////////////////////////////////
		// Logical literals: TRUE +1.0, FALSE -1.0, UNKNOWN 0.0 //
		//////////////////////////////////////////////////////////
		if (filler1 instanceof Logical && filler2 instanceof Logical){
			return compareDoubles
			       (logicalToDouble((Logical)filler1),
			        logicalToDouble((Logical)filler2));
		}
		//////////////////////////////////////////////////////
		// Unordered vectors: compare using subset relation //
		//////////////////////////////////////////////////////
		if (filler1 instanceof UnorderedVector
		 && filler2 instanceof UnorderedVector)
		{
			boolean gte= ((UnorderedVector)filler2).subsetOf
			             ((UnorderedVector)filler1, entityPair);
			boolean lte= ((UnorderedVector)filler1).subsetOf
			             ((UnorderedVector)filler2, entityPair);

			if (gte && lte) {
				return EQUAL;
			}
			if (gte) {
				return GREATER;
			}
			if (lte) {
				return LESS;
			}
			return INCOMPARABLE;
		}
		////////////////////////////////////////
		// Anything else: error message, exit //
		////////////////////////////////////////
		System.err.println("SemanticStrength.compareSSSfillers:"
		+ filler1 + " and " + filler2 + " are neither Frames nor "
		+ "Logicals nor UnorderedVectors!");
		System.exit(1); return 2;
	}

	/********************************************************************/
	// SemanticStrength.compareDoubles : (double x double) -> int
	/*********************************************************************
	** Compares two doubles, and returns GREATER if the first is greater
	** than the second, LESS if the first is less than the second, or
	** EQUAL if they are equal.
	** @param a The first double.
	** @param b The second double.
	** @return GREATER, LESS, or EQUAL.
	*********************************************************************/

	private static int compareDoubles (double a, double b) {
		if (a > b) {
			return GREATER;
		}
		if (a < b) {
			return LESS;
		}
		// (a == b)
		return EQUAL;
	}

	/********************************************************************/
	// SemanticStrength.logicalToDouble : Logical -> double
	/*********************************************************************
	** Converts a Logical to a double.
	** @param logical The Logical.
	** @return +1.0 if TRUE, -1.0 if FALSE, 0.0 if UNKNOWN.
	*********************************************************************/

	private static double logicalToDouble (Logical logical) {
		if (logical.equals(Logical.TRUE)) {
			return 1.0;
		}
		else if (logical.equals(Logical.FALSE)) {
			return -1.0;
		}
		else { //Logical.UNKNOWN
			return 0.0;
		}
	}
}
